module streams;
// version:0.02
/*
 * Stream interface.  Also includes a typical file stream implementation.
 *
 * Coded by Andy Friesen
 * 30 January 2004
 * Module renamed and modified by Larry Cowan
 *   changes: files now open in binary mode
 *            added 2 readChars functions to BinaryReader interface
 *            added filename function to FileStream class
 * 27 February 2004
 *
 * This is also something of a cool experiment on my part.  The whole thing
 * is based on what Matthew Wilson calls 'bolt-ins'.  Each component class
 * is a template that accepts a parent class as its argument.  This allows
 * for arbitrarily complex compositing of types and functionality, which is
 * very, very cool.
 *
 * Quick table of contents:
 *
 * Miscellany:
 *
 * Seek - enum of seek modes.  Pretty dull.
 * IOError - Exception that gets thrown when things go wrong.
 *
 * Interfaces:
 *
 * Readable - base interface for things that can be read from.  Implements
 *            a single raw input method.  More complex reading is done through
 *            the DataInput interface. (below)
 *            Also has the eof() method.
 * Writable - base writer interface.  Implements the basic write() method.
 *            Higher level functionality is provided by DataOutput. (below)
 * Seekable - base seeker interface.  Things that can seek implement this.
 *            More functionality is available through the Seeker class.
 *            (also below)
 * DataInput - interface providing higher level file input.  DataInput reads all
 *             of the D primitive types, as well as some string I/O.
 * DataOutput - interface providing high level file output.  DataOutput writes
 *              all of the D primitive types, in addition to a few string printing
 *              methods.
 *
 * Bolt-in classes:
 *
 * BinaryWriter - implements the DataOutput interface by writing things in some sort
 *                of hardware dependent binary form.
 * BinaryReader - implements the DataInput interface.  BinaryReader expects data in
 *                whatever format BinaryWriter writes.
 * TextWriter - implements DataOut, converting things to strings before
 *              writing to the stream.
 * Seeker - provides a slightly nicer interface for seeking.
 *          Maybe this should be removed as well.
 * FileStream - Base class for ReadableFileStream and WritableFileStream.
 *              FileStream manages the internal FILE* pointer, provides
 *              basic file manipulation methods (like close()), implements
 *              the Seekable interface, and is just really impressive looking
 *              overall.
 * ReadableFileStream - Extends FileStream.  Implements the Readable interface.
 * WritableFileStream - Also extends FileStream.  Implements the Writable interface.
 *
 * Actual classes:
 *
 * ReadableFile - implements Readable, Seekable, and DataInput.  Also incorporates
 *                Seeker.
 * WritableFile - Implements Writable, Seekable, and DataOutput.  Incorporates
 *                Seeker.
 *
 * Distribution rights:
 *      This code may be used freely for any purpose whatsoever, provided that
 *      the initial author(s) is/are never misrepresented in any way.
 */

private
{
    import std.c.stdio;
    import std.string;
}

// Different seek methods.
enum Seek
{
    // Seek from the start. (ie set the absolute file position)
    START = std.c.stdio.SEEK_SET,

    // Seek from the end
    END = std.c.stdio.SEEK_END,

    // Seek from the current position. (relative seek)
    CUR = std.c.stdio.SEEK_CUR
}

class IOError : Error
{
    this(char[] msg) { super(msg); }
}

/**
 * A stream that can be read from.
 *
 * This won't see much direct use either.
 */
interface Readable
{
    // returns the number of bytes actually read
    // throws IOError if an error occurs
    // EOF is not an error.
    int read(void* ptr, int len);

    bit eof();
}

/**
 * A writable stream.  der.
 */
interface Writable
{
    // returns the number of bytes written.
    // throws an IOError if an error occurs.
    int write(void* ptr, int len);
}

/**
 * Seekable stream!  golly!
 */
interface Seekable
{
    uint tell();
    void seek(long pos, Seek whence);
}

/**
 * Inspired by the Java interface of the same name.
 *
 * Pulls data from a binary stream, and returns it in some PDT or other.
 * There's also some handy string reading stuff here.
 */
interface DataInput
{
    bit readBit();
    byte readByte();
    ubyte readUByte();
    short readShort();
    ushort readUShort();
    int readInt();
    uint readUInt();
    long readLong();
    ulong readULong();
    // TODO: cent and ucent?
    float readFloat();
    double readDouble();
    real readReal();
    ireal readIReal();
    ifloat readIFloat();
    idouble readIDouble();
    creal readCReal();
    cfloat readCFloat();
    cdouble readCDouble();
    char readChar();
    wchar readWChar();
    dchar readDChar();
    // a handful of convenience methods:
    char[] readLine();
    byte[] readBytes(uint count);
    char[] readChars(uint count);
    // returns the number of bytes actually read
    int readBytes(byte[] buffer);
    int readChars(char[] buffer);
}

/**
 * Writes PDTs, byte[]s, and strings.
 */
interface DataOutput
{
    // Making the type part of the method name for the sake of symmetry.
    void writeBit    (bit     b);
    void writeByte   (byte    b);
    void writeUByte  (ubyte   b);
    void writeShort  (short   s);
    void writeUShort (ushort  s);
    void writeInt    (int     i);
    void writeUInt   (uint    i);
    void writeLong   (long    l);
    void writeULong  (ulong   l);
    // TODO: cent/ucent
    void writeFloat  (float   f);
    void writeDouble (double  d);
    void writeReal   (real    r);
    void writeIReal  (ireal   r);
    void writeIFloat (ifloat  f);
    void writeIDouble(idouble d);
    void writeCReal  (creal   r);
    void writeCFloat (cfloat  f);
    void writeCDouble(cdouble d);
    void writeChar   (char    c);
    // void writeWChar  (wchar   c);
    // void writeDChar  (dchar   c);

    void writeBytes (byte[] buffer);
    void writeString(char[] str);
    void writeLine  (char[] str);
}


/**
 * Implements DataInput.  Assumes that T implements Readable.
 * Except for characters and bytes, this reads from a binary
 * format that isn't necessarily consistent with anything
 * except for BinaryWriter. (below)
 */
class BinaryReader(T) : T, DataInput
{
    /*
     * Needed to implement Readable.
     * Must be overridden in a subclass.
     *
     * This has the problem of overriding the parent, if there
     * is one.  Not good.
     */
    bit eof() { assert(false); return 0; }
    int read(void* ptr, int len) { assert(false); return 0; }

    bit readBit()
    {
        return readByte() != 0;
    }

    byte readByte()
    {
        byte b;
        read(&b, byte.size);
        return b;
    }

    ubyte readUByte()
    {
        ubyte b;
        read(&b, ubyte.size);
        return b;
    }

    short readShort()
    {
        short s;
        read(&s, short.size);
        return s;
    }

    ushort readUShort()
    {
        ushort s;
        read(&s, ushort.size);
        return s;
    }

    int readInt()
    {
        int i;
        read(&i, int.size);
        return i;
    }

    uint readUInt()
    {
        uint i;
        read(&i, uint.size);
        return i;
    }

    long readLong()
    {
        long l;
        read(&l, long.size);
        return l;
    }

    ulong readULong()
    {
        ulong l;
        read(&l, ulong.size);
        return l;
    }

    // TODO: cent and ucent

    float readFloat()
    {
        float f;
        read(&f, float.size);
        return f;
    }

    double readDouble()
    {
        double d;
        read(&d, double.size);
        return d;
    }

    real readReal()
    {
        real r;
        read(&r, real.size);
        return r;
    }

    ireal readIReal()
    {
        ireal r;
        read(&r, ireal.size);
        return r;
    }

    ifloat readIFloat()
    {
        ifloat f;
        read(&f, ifloat.size);
        return f;
    }

    idouble readIDouble()
    {
        idouble d;
        read(&d, idouble.size);
        return d;
    }

    creal readCReal()
    {
        creal r;
        read(&r, creal.size);
        return r;
    }

    cfloat readCFloat()
    {
        cfloat f;
        read(&f, cfloat.size);
        return f;
    }

    cdouble readCDouble()
    {
        cdouble d;
        read(&d, cdouble.size);
        return d;
    }

    char readChar()
    {
        char c;
        read(&c, char.size);
        return c;
    }

    char[] readChars(uint count)
    {
        char[] buffer;
        buffer.length = count;
        int result = read(&buffer[0], count);
        return buffer[0 .. result];
    }

    // returns the number of bytes actually read
    int readChars(char[] buffer)
    {
        int result = read(&buffer[0], buffer.length);
        return result;
    }

    wchar readWChar()
    {
        wchar c;
        read(&c, wchar.size);
        return c;
    }

    dchar readDChar()
    {
        dchar c;
        read(&c, dchar.size);
        return c;
    }

    // a handful of convenience methods:
    byte[] readBytes(uint count)
    {
        byte[] buffer;
        buffer.length = count;
        int result = read(&buffer[0], count);
        return buffer[0 .. result];
    }

    // returns the number of bytes actually read
    int readBytes(byte[] buffer)
    {
        int result = read(&buffer[0], buffer.length);
        return result;
    }

    char[] readLine()
    {
        char[] str;

        while (true)
        {
            char c = readChar();
            debug printf("c='%c'\n", c);
            if (c == '\n' || eof())
                return str;
            else
                str ~= c;
        }
    }
}

/**
 * Wrapper that implements DataInput by reading from a stream.
 * This is handy when you have an existing stream class, and need
 * to pull data from it, even though it does not itself implement
 * DataInput.
 */
class StreamReadAdapter : BinaryReader!(Object)
{
    this(Readable stream)
    {
        _stream = stream;
    }

    int read(void* dest, uint length)
    {
        return _stream.read(dest, length);
    }

private:
    Readable _stream;
}

/**
 * Writes PDTs, byte[]s, and strings in some scary,
 * hardware dependant manner that is consistent with BinaryReader,
 * but not necessarily anything else.
 *
 * The more I think about it, the more I get to thinking that we need a
 * DataOut implementation that converts everything to a string before
 * writing it.
 */
class BinaryWriter(T) : T, DataOutput
{
    // implementing Writable
    // has the same problems as BinaryReader.read
    int write(void* ptr, int len) { assert(false); return 0; }

    void writeBit    (bit     b) { writeByte(b ? 1 : 0);    }
    void writeByte   (byte    b) { write(&b, byte   .size); }
    void writeUByte  (ubyte   b) { write(&b, ubyte  .size); }
    void writeShort  (short   s) { write(&s, short  .size); }
    void writeUShort (ushort  s) { write(&s, ushort .size); }
    void writeInt    (int     i) { write(&i, int    .size); }
    void writeUInt   (uint    i) { write(&i, uint   .size); }
    void writeLong   (long    l) { write(&l, long   .size); }
    void writeULong  (ulong   l) { write(&l, ulong  .size); }
    // TODO: cent/ucent
    void writeFloat  (float   f) { write(&f, float  .size); }
    void writeDouble (double  d) { write(&d, double .size); }
    void writeReal   (real    r) { write(&r, real   .size); }
    void writeIReal  (ireal   r) { write(&r, ireal  .size); }
    void writeIFloat (ifloat  f) { write(&f, ifloat .size); }
    void writeIDouble(idouble d) { write(&d, idouble.size); }
    void writeCReal  (creal   r) { write(&r, creal  .size); }
    void writeCFloat (cfloat  f) { write(&f, cfloat .size); }
    void writeCDouble(cdouble d) { write(&d, cdouble.size); }
    void writeChar   (char    c) { write(&c, char   .size); }
    // void writeWChar  (wchar   c) { write(&c, wchar  .size); }
    // void writeDChar  (dchar   c) { write(&c, dchar  .size); }

    void writeBytes (byte[] buffer)
    {
        if (buffer.length)
            write(&buffer[0], buffer.length);
    }

    void writeString(char[] str)
    {
        if (str.length)
            write(&str[0], str.length);
    }

    void writeLine(char[] str)
    {
        writeString(str);
        writeChar('\n');
    }
}

/**
 * Same idea as BinaryWriter, but converts everything to a string
 * first.
 */
class TextWriter(T) : T, DataOutput
{
    // implementing Writable
    // has the same problems as BinaryReader.read
    int write(void* ptr, int len) { assert(false); return 0; }

    void writeBit    (bit     b) { writeString(b ? "true" : "false");     }
    void writeByte   (byte    b) { writeString(.toString(b)); }
    void writeUByte  (ubyte   b) { writeString(.toString(b)); }
    void writeShort  (short   s) { writeString(.toString(s)); }
    void writeUShort (ushort  s) { writeString(.toString(s)); }
    void writeInt    (int     i) { writeString(.toString(i)); }
    void writeUInt   (uint    i) { writeString(.toString(i)); }
    void writeLong   (long    l) { writeString(.toString(l)); }
    void writeULong  (ulong   l) { writeString(.toString(l)); }
    // TODO: cent/ucent
    void writeFloat  (float   f) { writeString(.toString(f)); }
    void writeDouble (double  d) { writeString(.toString(d)); }
    void writeReal   (real    r) { writeString(.toString(r)); }
    void writeIReal  (ireal   r) { writeString(.toString(r)); }
    void writeIFloat (ifloat  f) { writeString(.toString(f)); }
    void writeIDouble(idouble d) { writeString(.toString(d)); }
    void writeCReal  (creal   r) { writeString(.toString(r)); }
    void writeCFloat (cfloat  f) { writeString(.toString(f)); }
    void writeCDouble(cdouble d) { writeString(.toString(d)); }
    void writeChar   (char    c) { write(&c, char.size);    }

    // these won't work right.  TODO: fix
    // void writeWChar  (wchar   c) { writeString(.toString(c)); }
    // void writeDChar  (dchar   c) { writeString(.toString(c)); }

    /**
     * this is kind of weird.  It's here because DataOutput
     * says it has to be. >_>
     */
    void writeBytes (byte[] buffer)
    {
        if (buffer.length)
            write(&buffer[0], buffer.length);
    }

    void writeString(char[] str)
    {
        if (str.length)
            write(&str[0], str.length);
    }

    void writeLine(char[] str)
    {
        writeString(str);
        writeChar('\n');
    }
}

/**
 * Mirror image of StreamReadAdapter.  Accepts a Writable
 * interface, and wraps it, providing the DataOutput interface.
 */
class StreamWriteAdapter : BinaryWriter!(Object)
{
    this(Writable stream)
    {
        _stream = stream;
    }

    int write(void* ptr, uint size)
    {
        return _stream.write(ptr, size);
    }

private:
    Writable _stream;
}

/**
 * Convenience seeking methods.
 *
 * overkill.  But this design is kind of cool.
 */
class Seeker(T) : T
{
    // These should be overriden by child classes
    uint tell() { assert(false); return 0;}
    void seek(long pos, Seek whence) { assert(false); }

    void seekAbsolute(long pos)
    {
        seek(pos, Seek.START);
    }

    void seekFromEnd(long pos)
    {
        seek(pos, Seek.END);
    }

    void seekRelative(long pos)
    {
        seek(pos, Seek.CUR);
    }
}

/**
 * Base class for file streams.
 *
 * All files are seekable.
 */
class FileStream(T) : T, Seekable
{
    bit isOpen()
    {
        return handle != null;
    }

    void close()
    {
        if (_handle !== null)
            fclose(_handle);
        _handle = null;
    }

    ~this()
    {
        close();
    }

    // Seekable
    uint tell()
    {
        return std.c.stdio.ftell(handle);
    }

    void seek(long pos, Seek whence)
    {
        std.c.stdio.fseek(handle, pos, cast(int)whence);
    }

    char[] filename()
    {
        return _fileName;
    }

protected:
    this(FILE* handle)
    {
        _handle = handle;
    }

    FILE* handle()
    {
        return _handle;
    }

    FILE* _handle;

    void setFileName ( char[] fileName )
    {
        _fileName = fileName;
    }

    char[] _fileName;

}

/**
 * Readable file stream
 *
 * Implementation classes like this need to be put near the
 * end of the heirarchy, as the helper classes all have to implement
 * Readable, Writable, and so forth.  As a result, they override the
 * relevant methods to assert(false).
 *
 * Implementations need to be at the bottom of the heirarchy so they
 * override the assert(false) methods, and not the other way around.
 */
class ReadableFileStream(T) : FileStream!(T), Readable
{
    this(char[] fileName)
    {
        setFileName(fileName);
        FILE* f = std.c.stdio.fopen(toStringz(fileName), "rb");
        if (!f)
            throw new IOError("Unable to open " ~ fileName);

        super(f);
    }

    /**
     * This allows ReadableFileStream to inherit WritableFileStream,
     * or vice versa.
     */
    protected this(FILE* f)
    {
        super(f);
    }

    // Readable
    bit eof()
    {
        return std.c.stdio.feof(handle) != 0;
    }

    int read(void* buffer, int count)
    {
        return std.c.stdio.fread(buffer, 1, count, handle);
    }

}

class WritableFileStream(T) : FileStream!(T), Writable
{
    this(char[] fileName)
    {
        setFileName(fileName);
        FILE* f = std.c.stdio.fopen(toStringz(fileName), "wb");
        if (!f)
            throw new IOError("Unable to open " ~ fileName);

        super(f);
    }

    protected this(FILE* f)
    {
        super(f);
    }

    // Writable
    int write(void* buffer, int count)
    {
        return std.c.stdio.fwrite(buffer, 1, count, handle);
    }
}

class ConsoleWriter(T) : WritableFileStream!(T)
{
    this()
    {
        super(std.c.stdio.stdout);
    }
}

// Aliases for sane type names:

alias ReadableFileStream!(
    BinaryReader!(
    Seeker!(Object)
    )) ReadableFile;

alias WritableFileStream!(
    BinaryWriter!(
    Seeker!(Object)
    )) WritableFile;

alias ConsoleWriter!(
    TextWriter!(
    Object
    )) ConsoleStream;

ConsoleStream Console;

static this()
{
    Console = new ConsoleStream();
}

