module wildmatch;
// version:0.02
/+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 + Coded by Larry Cowan
 + March 8, 2004
 + March 17, 2004 lfc - state machine replaced, supporting [abc] and [^abc]
 +
 + This class stores a unix-style wildcard pattern in an instance and then
 + can test filenames for matching the pattern. * (zero or more of any
 + character) and ? (any one character) and [abc] and [^xyz] are supported
 + in wildcard content.
 + This matching is case-insensitive for win32 and case-sensitive for any
 + other platform - or it may be specified to the constructor.
 +
 +++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
 +  Copyright 2004 by Larry F. Cowan
 +  Distribution rights:  This code may be used for any purpose whatever,
 +      commercial or otherwise, without specific permission from the 
 +      author(s), as long as they are not misrepresented in any way, the
 +      author name(s) are kept in the source, and this notice is included.
 +      Changes should indicate at least an author name and date.  No listed
 +      author takes on any responsibility or liability for your use of this
 +      module. It's free.  Caveat emptor!  Any underlying (imported) modules
 +      are under their own licensing and this is not affected here. 
 ++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++/
//debug=Wildmatch;  // compile test main()
//debug=wildmatch;  // compile debug messages
version=win32;

class WildMatch
{
    private char[] pattern;
    private int[] pstates;
    private int plen;
    private bit case_insensitive;

    this ( char[] initpattern ) {
        case_insensitive = false;
        version (win32) case_insensitive = true;
        pattern = (initpattern.length > 0) ? initpattern : "*";
        if (case_insensitive)
            foreach( int i, char c; pattern )
                pattern[i] = setupper(c);
        plen = pattern.length;
        debug (wildmatch) printf("Init(%d): \"%.*s\"\n",plen,pattern);
    }

    this ( char[] initpattern, char[] options ) {
        case_insensitive = false;
        foreach ( char opt; options )
            if ( opt == 'i' )
                case_insensitive = true;
        pattern = (initpattern.length > 0) ? initpattern : "*";
        if (case_insensitive)
            foreach( int i, char c; pattern )
                pattern[i] = setupper(c);
        plen = pattern.length;
        debug (wildmatch) printf("Init(%d): \"%.*s\"\n",plen,pattern);
    }

    private char setupper(char c)
    {
        return (c < 'a' || c > 'z') ? c : c - ('a' - 'A');
    }

    private void trace ( int i, char c, int q )
    {
        if (q < plen)
            printf("%d%c. setting pattern%d[%c] %.*s\n",i,c,q,pattern[q],(pstates[q]<1)?"off":"on");
        else
            printf("%d%c. setting pattern%d[plen] %.*s\n",i,c,q,(pstates[q]<1)?"off":"on");
    }

    debug (Wildmatch) private char[] w () { return pattern; }

    public bit test ( char[] filename )
    in {
        assert( filename.length > 0 );
    }
    body {
        bit caret;
        bit tempok;
        int i, j;

        debug (wildmatch) printf("\nPattern(%d): \"%.*s\" vs \"%.*s\"\n",
                                 plen,pattern,filename);
        pstates.length = plen + 1;
        for (i=0; i < plen + 1 ;i++)
            pstates[i] = 0;
        if ( pattern[i=0] == '*' ) {
            for (j=1; j < plen ;j++)
                if ( pattern[j] != '*' ) // scan successive stars
                     break;
            if ( j == plen - 1 )
                return true;   // short-circuit = pattern is all stars
            pstates[j - 1] = 1; // turn on last of group of stars
            debug (wildmatch) trace(i,'a',j-1);
            pstates[j] = 1;    // accept - turn on following character
            debug (wildmatch) trace(i,'a',j);
        } else {
            pstates[i=0] = 1;
            debug (wildmatch) trace(i,'a',0);
        }
        foreach ( char ch; filename ) {
            debug (wildmatch) printf("[%c]:\n",ch);
            pstates[plen] = 0;
            debug (wildmatch) trace(i,'b',plen);
            for (i=plen-1; i >= 0 ;i--) {
                if ( pstates[i] == 0 ) // if state inactive, ignore it
                    continue;
                switch ( pattern[i] ) {
                  case '*':
                    if ( i == plen - 1 )
                        return true;   // short-circuit = pattern ends in '*'
                    for (j=i+1; j < plen ;j++)
                        if ( pattern[j] != '*' ) // scan successive stars
                             break;
                    pstates[i] = 0;     // turn off current star 
                    debug (wildmatch) trace(i,'c',i);
                    pstates[j - 1] = 1; // turn on last of group of stars
                    debug (wildmatch) trace(i,'d',j-1);
                    pstates[j] = 1;    // accept - turn on following character
                    debug (wildmatch) trace(i,'e',j);
                    break;
                  case '?':
                    pstates[i] = 0;    // accept - move to next char
                    debug (wildmatch) trace(i,'f',i);
                    pstates[i+1] = 1;  //
                    debug (wildmatch) trace(i,'g',i+1);
                    break;
                  case ']':
                    return false;  // short-circuit - missing ']'
                  case '[':
                    j = i;
                    if (pattern[j+1] == '^') {
                        caret = true;
                        j++;
                    } else
                        caret = false;
                    for (tempok=caret,j++; pattern[j] != ']' ;j++) {
                        if ( j == plen )
                            return false;  // short-circuit - unclosed '['
                        if ( pattern[j] == setupper(ch) )
                            if ( caret )
                                tempok = false;
                            else
                                tempok = true;
                    }
                    if ( tempok ) {
                        pstates[i] = 0;  // accept, go to next char in pattern
                        debug (wildmatch) trace(i,'h',i);
                        pstates[j+1] = 1;
                        debug (wildmatch) trace(i,'i',j+1);
                    } else {
                        pstates[i] = 0; // [] failed
                        debug (wildmatch) trace(i,'j',i);
                    }
                    break;
                  default:
                    if ( ch == pattern[i] 
                     || (case_insensitive && setupper(ch) == pattern[i])) {
                        pstates[i] = 0;  // accept, move to next char in patt
                        debug (wildmatch) trace(i,'k',i);
                        pstates[i+1] = 1;
                        debug (wildmatch) trace(i,'l',i+1);
                    } else {
                        pstates[i] = 0;  // ch failed
                        debug (wildmatch) trace(i,'m',i);
                    }
                    break;
                }
            }
            for (i=plen-1; i >= 0 ;i--) { // post-process any active stars
                if ( pstates[i] == 0      // in case we're done with filename
                  || pattern[i] != '*' )
                    continue;
                for (j=i+1; j < plen ;j++)
                    if ( pattern[j] != '*' ) // scan successive stars
                         break;
                pstates[j] = 1;    // turn on following character
                debug (wildmatch) trace(i,'n',j);
            }
        }
        if ( pstates[plen] > 0 ) {
            debug (wildmatch) printf("TRUE\n");
            return true;  // end of pattern coincides with end of filename
        }
        debug (wildmatch) printf("FALSE\n");
        return false;     // else failed
    }
}

/* ****************************************** *
 *       main() - used for testing only       *
 * ****************************************** */
debug (Wildmatch) {
  void main ( char[][] args )
  {
    if (args.length != 2 && args.length != 3) {
        printf("Usage: filename [pattern]\n");
        return;
    }
    char[] fn = args[1];

    if (args.length == 3) {
        WildMatch wc0 = new WildMatch(args[2]);
        printf("wc0(%.*s) is %.*s\n",wc0.w(),wc0.test(fn) ? "true" : "false");
        return;
    }

    WildMatch wc1 = new WildMatch("wildmatch.d");
    WildMatch wc2 = new WildMatch("wildmatch.d*");
    WildMatch wc3 = new WildMatch("w?ld??*.?");
    WildMatch wc4 = new WildMatch("*");
    WildMatch wc5 = new WildMatch("*.*");
    WildMatch wc6 = new WildMatch("*.d");
    WildMatch wc7 = new WildMatch("s*");
    WildMatch wc8 = new WildMatch("s*.*");
    WildMatch wc9 = new WildMatch("s*.d");
    WildMatch wc10 = new WildMatch("?*");
    WildMatch wc11 = new WildMatch("*?");
    WildMatch wc12 = new WildMatch("**");
    WildMatch wc13 = new WildMatch("??");
    WildMatch wc14 = new WildMatch("*********");
    WildMatch wc15 = new WildMatch("*?**?***?****?*****");
    WildMatch wc16 = new WildMatch("wild[jklmn][aeiou]tch.[cd]");
    WildMatch wc17 = new WildMatch("wild[^jkln][^eiou]tch.[^d]xe");

    printf("wc1(%.*s) is %.*s\n",wc1.w(),wc1.test(fn) ? "true" : "false");
    printf("wc2(%.*s) is %.*s\n", wc2.w(), wc2.test(fn) ? "true" : "false");
    printf("wc3(%.*s) is %.*s\n", wc3.w(), wc3.test(fn) ? "true" : "false");
    printf("wc4(%.*s) is %.*s\n", wc4.w(), wc4.test(fn) ? "true" : "false");
    printf("wc5(%.*s) is %.*s\n", wc5.w(), wc5.test(fn) ? "true" : "false");
    printf("wc6(%.*s) is %.*s\n", wc6.w(), wc6.test(fn) ? "true" : "false");
    printf("wc7(%.*s) is %.*s\n", wc7.w(), wc7.test(fn) ? "true" : "false");
    printf("wc8(%.*s) is %.*s\n", wc8.w(), wc8.test(fn) ? "true" : "false");
    printf("wc9(%.*s) is %.*s\n", wc9.w(), wc9.test(fn) ? "true" : "false");
    printf("wc10(%.*s) is %.*s\n", wc10.w(), wc10.test(fn) ? "true" : "false");
    printf("wc11(%.*s) is %.*s\n", wc11.w(), wc11.test(fn) ? "true" : "false");
    printf("wc12(%.*s) is %.*s\n", wc12.w(), wc12.test(fn) ? "true" : "false");
    printf("wc13(%.*s) is %.*s\n", wc13.w(), wc13.test(fn) ? "true" : "false");
    printf("wc14(%.*s) is %.*s\n", wc14.w(), wc14.test(fn) ? "true" : "false");
    printf("wc15(%.*s) is %.*s\n", wc15.w(), wc15.test(fn) ? "true" : "false");
    printf("wc16(%.*s) is %.*s\n", wc16.w(), wc16.test(fn) ? "true" : "false");
    printf("wc17(%.*s) is %.*s\n", wc17.w(), wc17.test(fn) ? "true" : "false");
  }
}
